// muesli-octave.cpp -*- C++ -*-
/* Interface to octave evaluators / template for new language interfaces
   Copyright (C) 2008, 2009, 2010 University of Limerick

   This file is part of Muesli.
   
   Muesli is free software: you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation, either version 3 of the License, or (at your
   option) any later version.
   
   Muesli is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
   
   You should have received a copy of the GNU General Public License
   along with Muesli.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _MUESLI_OCTAVE_CPP_
#define _MUESLI_OCTAVE_CPP_

extern "C" {
#include "../config.h"
}

#ifdef HAVE_LIBOCTINTERP

extern "C" {
#include "muesli.h"
}

// We've defined these in our config.h; clear them to stop warnings
// from octave's config.h redefining them:
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION

#include "octave/config.h"
#include "octave/octave.h"
#include "octave/symtab.h"
#include "octave/parse.h"
#include "octave/unwind-prot.h"
#include "octave/toplev.h"
#include "octave/error.h" 
#include "octave/quit.h"
#include "octave/variables.h"
#include "octave/sighandlers.h"
#include "octave/sysdep.h"
#include "octave/version.h"

using namespace std;

extern "C" {
#include <stdio.h>
#include <limits.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>		// todo: do I need this?

  void octave_set_parameter();
  void octave_set_parameters();
  void octave_get_parameter();
  void octave_get_parameters();
  void muesli_octave_eval_in_language();
  void octave_eval_custom();
  void octave_evaluator_init(evaluator_interface *interface);
  void octave_load_file(evaluator_interface *interface,
			const char *filename);
  result_type_t octave_eval_string(evaluator_interface *evaluator,
				   const char *octave_c_string,
				   unsigned int string_length,
				   eval_result_t *result_p,
				   int transient);
  void octave_setup(evaluator_interface *new_octave_interface);
}

// #include "muesli-octave.h"

static int ambient_transient = 0;

typedef struct octave_state {
  char *dummy;
} octave_state;

#ifndef CUSTOMIZED
#if 0
octave_state *octave_create_octave() { return NULL; }

// Getting args from your scripting language stack:
int octave_count_args(octave_state *S) { return 0; }
int octave_number_arg(octave_state *S, int argi) { return 0; }
char *octave_string_arg(octave_state *S, int argi) { return (char*)""; }
int octave_arg_is_string(octave_state *S, int argi) { return 0; }
int octave_arg_is_number(octave_state *S, int argi) { return 0; }

// Putting results back onto your scripting language stack:
void octave_return_string(octave_state *S, const char *string_back) {}
void octave_return_number(octave_state *S, float number_back) {}
void octave_return_boolean(octave_state *S, int int_back) {}
void octave_return_void(octave_state *S) {};

// Handling table / map / alist structures of your scripting language:
octave_table *octave_create_table(octave_state *S, int size) { return NULL; }
void octave_set_table_string(octave_state *S, octave_table *T, char *key, const char *value) {}
void octave_set_table_number(octave_state *S, octave_table *T, char *key, float value) {}
void octave_set_table_boolean(octave_state *S, octave_table *T, char *key, int value) {}

// Make a name / symbol of the scripting language:
octave_symbol *octave_make_symbol(octave_state *S, const char *n) { return NULL; }

// Non-local exits in the scripting language
void octave_throw(octave_state *S, octave_symbol *Y, float n) {}

// Set a global variable of the scripting language to contain a table:
void octave_set_global_variable_table(octave_state *S, octave_symbol *varname, octave_table *table) {}

// Extend the scripting language with an added built-in function:
void octave_register_builtin_function(octave_state *S, const char *funname, void (*function)(octave_state *FS)) {}

// A specialized non-local exit; could either throw, or just quit the
// application run:
void
octave_report_error(octave_state *S, char *message, char *arg)
{
  fprintf(stderr, "Got error \"%s\", \"%s\"\n", message, arg);
}
#endif
#endif

static evaluator_interface *octave_interface = NULL;

// static char *option_names_var_name = (char*)"option_names";

static unsigned int scratch_len = 0;
static char* scratch_buf = NULL;

void
octave_set_parameter()
{
#if 0
  if (octave_count_args(cs) < 2) {
    fprintf(stderr, "too few args to octave_set_parameter\n");
    octave_error(cs, (char*)"argcount", (char*)"set_parameter");
  }

  char option_code = muesli_find_option_letter(octave_interface->getopt_options, octave_string_arg(cs, 0));

  if (option_code != -1) {
    eval_result_t result;
    result_type_t result_type;
    if (octave_arg_is_string(cs, 1)) {
      const char *from_octave = octave_string_arg(cs, 1);
      if (strlen(from_octave)+1 > scratch_len) {
	scratch_len = strlen(from_octave)+1;
	if (scratch_buf != NULL) {
	  free(scratch_buf);
	}
	scratch_buf = (char*)malloc(scratch_len);
      }
      strcpy(scratch_buf, from_octave);
      (octave_interface->handle_option)(octave_interface->app_params,
					option_code, scratch_buf, 0.0, 1,
					&result_type, &result, "octave");
    } else if (octave_arg_is_number(cs, 1)) {
      (octave_interface->handle_option)(octave_interface->app_params,
					option_code, NULL, octave_number_arg(cs, 2), 1,
					&result_type, &result, "octave");
    } else {
      (octave_interface->handle_option)(octave_interface->app_params,
					option_code, (char*)"true", 0.0, 1,
					&result_type, &result, "octave");
    }
  }
#endif
}

void
octave_set_parameters()
{
#if 0
  if (octave_count_args(cs) < 1)  {
    fprintf(stderr, "too few args to octave_set_parameters\n");
    octave_error(cs, (char*)"argcount", (char*)"set_parameters");
  }

#if 0
  // Fill in: use a table iterator from your language
  octave_table table = octave_table_arg(cs, 0);

  octave_table_iteration_start(cs, table);
  while (octave_table_iteration_next(cs, table) != 0) {
    octave_set_parameter(cs,
			 octave_table_iteration_current_key(cs, table),
			 octave_table_iteration_current_value(cs, table));
  }

#endif
#endif
}

void
octave_get_parameter()
{
#if 0
  if (octave_count_args(cs) < 1)  {
    fprintf(stderr, "too few args to octave_get_parameter\n");
    octave_error(cs, (char*)"argcount", (char*)"get_parameter");
  }

  char option_code = muesli_find_option_letter(octave_interface->getopt_options,
					       octave_string_arg(cs, 0));

  if (option_code != -1) {

    result_type_t result_type;
    eval_result_t result;

    (octave_interface->handle_option)(octave_interface->app_params,
				      option_code,	// option
				      NULL, 0.0,	// value
				      0,	// set
				      &result_type, &result,
				      "octave");

    switch (result_type) {
    case string_result:
      octave_return_string(cs, result.as_string);
      break;
    case number_result:
      octave_return_number(cs, result.as_number);
      break;
    case boolean_result:
      octave_return_boolean(cs, result.as_int);
      break;
    default:
      octave_return_boolean(cs, 0);
      break;
    }
  }
#endif
}

void
octave_get_parameters()
{
#if 0
  octave_table *table = octave_create_table(cs, 48);

  // todo: fix and re-instate -- I have to get long_options across to it somehow
  struct option *option_names = octave_interface->getopt_options;

  while (option_names->name != 0) {
    
    result_type_t result_type;
    eval_result_t result;

    (octave_interface->handle_option)(octave_interface->app_params,
				      (option_names->val), // opt
				      NULL, 0.0,		     // value
				      0,		     // set
				      &result_type, &result,
				      "octave");

    switch (result_type) {
    case string_result:
      octave_set_table_string(cs, table, (char*)(option_names->name), result.as_string);
      break;
    case number_result:
      octave_set_table_number(cs, table, (char*)(option_names->name), result.as_number);
      break;
    case boolean_result:
      octave_set_table_boolean(cs, table, (char*)(option_names->name), result.as_int);
      break;
    default:
      octave_set_table_boolean(cs, table, (char*)(option_names->name), 0);
      break;
    }
    option_names++;
  }
#endif
}

///////////////////////////////////////
// Call arbitrary evaluators by name //
///////////////////////////////////////

void
muesli_octave_eval_in_language()
{
#if 0
  if (octave_count_args(cs) < 2) {
    fprintf(stderr, "too few args to octave_eval_in_language\n");
    octave_error(cs, (char*)"argcount", (char*)"eval_in_language");
    return 0;
  }

  if ((octave_arg_is_string(cs, 1)) && (octave_arg_is_string(cs, 2))) {
    const char *language_name = octave_string_arg(cs, 1);
    const char *evaluand = octave_string_arg(cs, 2);
    unsigned int evaluand_length = strlen(evaluand);

    fprintf(stderr, "In octave_eval_in_language(\"%s\", \"%s\")\n", language_name, evaluand);

    eval_result_t result;

    result_type_t result_type = muesli_eval_in_language(language_name,
							evaluand,
							evaluand_length,
							&result,
							ambient_transient);

    switch (result_type) {
    case number_result:
      octave_return_number(cs, result.as_number);
      break;
    case string_result:
      octave_return_string(cs, result.as_string);
      break;
    case boolean_result:
      octave_return_boolean(cs, result.as_int);
      break;
    case no_result:
    case error_result:
      octave_return_boolean(cs, 0);
      break;
    }
  } else {
    fprintf(stderr, "wrong type args to octave_eval_in_language\n");
    octave_error(cs, (char*)"argtype", (char*)"eval_in_language");
    return 0;
  }
  return 1;
#endif
}

///////////////////////////////
// Octave built-in functions //
///////////////////////////////

// This isn't meant as self-referentially as it looks (pity!).  It's
// really there so that if you use this file as a template for another
// interpreter interface, the language you're adding will get access
// to octave built-in functions just like the existing languages do.

void
octave_eval_custom()
{
#if 0
  if (octave_arg_is_string(cs, 0)) {
    char *string_arg = octave_string_arg(cs, 0);
    eval_result_t result;
    result_type_t result_type = (octave_interface->eval_string)(octave_interface,
								string_arg, strlen(string_arg),
								&result, ambient_transient);
    switch (result_type) {
    case number_result:
      octave_return_number(cs, result.as_number);
      break;
    case string_result:
      octave_return_string(cs, result.as_string);
      break;
    case boolean_result:
      octave_return_boolean(cs, result.as_int);
      break;
    case no_result:
    case error_result:
      octave_return_void(cs);
      break;
    }
  } else {
    fprintf(stderr, "octave_eval_function must be given a string\n");
    octave_error(cs, (char*)"argtype", (char*)"octave_eval");
  }
#endif
}

////////////////
// Initialize //
////////////////

void
octave_evaluator_init(evaluator_interface *interface)
{
  // Init octave interface

  char *oargv[3];
  int oargc = 2;

  oargv[0] = (char*)"libmuesli";
  oargv[1] = (char*)"--silent";
  oargv[2] = NULL;

  octave_state *our_octave_state = NULL;

  octave_main(oargc,oargv,1);

  interface->state = our_octave_state;
  octave_interface = interface;

  // Extend system

  // Fill in: register all these language extensions (as applicable):
  Muesli_Add_Fn_1(interface, (char*)"get_parameter", octave_get_parameter);
  Muesli_Add_Fn_2(interface, (char*)"set_parameter", octave_set_parameter);
  Muesli_Add_Fn_0(interface, (char*)"evolution_parameters", octave_get_parameters);
  Muesli_Add_Fn_1(interface, (char*)"modify_evolution_parameters", octave_set_parameters);
  Muesli_Add_Fn_1(interface, (char*)"custom", octave_eval_custom);

  // Set up option names
#if 0
  // Fill in: Create a table of option names.  You may well not need to bother.
  octave_table *option_names_table = octave_create_table(our_octave_state, 48);
  struct option *option_names;
  for (option_names = octave_interface->getopt_options;
       option_names->name != 0;
       option_names++) {
    octave_set_table_number(our_octave_state, option_names_table,
			    (char*)(option_names->name), (option_names->val));
  }
  octave_set_global_variable_table(our_octave_state,
				   octave_make_symbol(our_octave_state,
						      option_names_var_name),
				   option_names_table);
#endif
}

void
octave_load_file(evaluator_interface *interface,
		const char *filename)
{
  int muesli_flags = interface->flags;
  if (muesli_flags & TRACE_MUESLI_LOAD) {
    fprintf(stderr, "Loading %s\n", filename);
  }

  source_file(filename, "base");

  if (muesli_flags & TRACE_MUESLI_LOAD) {
    fprintf(stderr, "Loaded %s\n", filename);
  }
}

result_type_t
octave_eval_string(evaluator_interface *evaluator,
		   const char *octave_c_string,
		   unsigned int string_length,
		   eval_result_t *result_p,
		   int transient)
{
  result_type_t result_type = no_result;

  if (octave_c_string) {


    int old_ambient_transient = ambient_transient;
    ambient_transient = transient;
    // fprintf(stderr, "Octave evaluating string \"%s\"\n", octave_c_string);

    octave_value_list raw_results;

    // Modified slightly from example code on the Octave web site,
    // which in turn is derived from toplev.cc which is part of Octave
    // (GPL3):

    int parse_status;

    octave_save_signal_mask ();
    if (octave_set_current_context) {
#if defined (USE_EXCEPTIONS_FOR_INTERRUPTS)
	panic_impossible ();
#else
	unwind_protect::run_all ();
	raw_mode (0);
	std::cout << "\n";
	octave_restore_signal_mask ();
#endif
      }

    can_interrupt = true;
    octave_catch_interrupts ();
    octave_initialized = true;

    // XXX FIXME XXX need to give caller an opaque pointer
    // so that they can define and use separate namespaces
    // when calling octave, with a shared global namespace.
    // Something like:
    //   int call_octave (const char *string, void *psymtab = NULL) {
    //     ...
    //     curr_sym_tab = psymtab == NULL ? top_level_sym_tab : symbol_table;
    // I suppose to be safe from callbacks (surely we have to
    // provide some way to call back from embedded octave into
    // the user's application), we should push and pop the current
    // symbol table.

    // Note that I'm trying to distinguish exception from 
    // failure in the return codes. I believe failure is 
    // indicated by -1.  I have execution exception (including
    // user interrupt and more dramatic failures) returning -2
    // and memory failure returning -3.  We should formalize
    // this with error codes defined in embed_octave.h.  Maybe
    // a more fine-grained approach could be used within octave
    // proper.
    try {
	curr_sym_tab = top_level_sym_tab;
	reset_error_handler ();
	raw_results = eval_string(octave_c_string, false, parse_status);
      }
    catch (octave_interrupt_exception) {
#if 0
    // todo: re-instate this block and find how to get the definition of recover_from_exception, which is in toplev.h which we already include
	recover_from_exception ();
#endif
	std::cout << "\n"; 
	error_state = -2; 
      }
    catch (std::bad_alloc) {
#if 0
    // todo: re-instate this block and find how to get the definition of recover_from_exception, which is in toplev.h which we already include
	recover_from_exception ();
#endif
	std::cout << "\n"; 
	error_state = -3;
      }
    octave_restore_signal_mask();
    octave_initialized = false;

    if (raw_results.empty()) {
      result_type = no_result;
    } else {

      octave_value res = raw_results(0);


      if (res.is_undefined() || res.is_empty()) {
	result_type = no_result;
      } else if (res.is_real_scalar()) {
	(*result_p).as_number = res.scalar_value(); 
	result_type = number_result;
      } else if (res.is_bool_scalar()) {
	(*result_p).as_bool = res.bool_value();
	result_type = boolean_result;
      } else if (res.is_string()) {
	(*result_p).as_string = res.string_value().c_str();
	result_type = string_result;
      } else {
	result_type = no_result;
      }




    
    }

    // delete raw_results;


    ambient_transient = old_ambient_transient;
  }

  return result_type;
}

void
octave_setup(evaluator_interface *new_octave_interface)
{
  octave_interface = new_octave_interface;

  octave_interface->eval_string = octave_eval_string;
  octave_interface->load_file = octave_load_file;

  octave_interface->version = (char*)OCTAVE_VERSION;
}

#endif
#endif

